#if !defined(TFTP_TEST_HARNESS)
  #include <common.h>
  #include <exports.h>
  #include <net.h>
  #include <nand.h>
  #include <jffs2/jffs2.h>
  #include <stdio_dev.h>
  #include <asm/errno.h>
#else
  #include "tftpsrv_harness.h"
#endif

#include "tftpsrv.h"

#if defined(CONFIG_CMD_NET)

// define these macros to either 1 or 0  to enable/disable features
#define Enable_TFTP_Server     1
#define EnableFakeMemoryDriver 0
#define EnableRamDriver        1
#define EnableNandDriver       1

#if !Enable_TFTP_Server
// tftp server not enabled

// Because tftp command will never be registered (at bottom of file),
// this function won't ever be called. But, the U-Boot network stack
// is a bit quirky in that it has to have hard-coded calls to "start"
// any protocol that can be supported.
void TftpServerStart (void)
{
}

#else
// tftp server is enabled

#if defined(TFTP_TEST_HARNESS)
  #define ListenPort 9969
#else
  #define ListenPort 69
#endif


// TFTP opcodes

#define TFTP_READ_REQ	1
#define TFTP_WRITE_REQ	2
#define TFTP_DATA	3
#define TFTP_ACK	4
#define TFTP_ERROR	5


#define TIMEOUT 1     // length of rx timeout in seconds
#define TIMEOUTMAX 5  // number of timeouts before transfer is aborted

static  unsigned captureIndex;


#if defined(TFTP_TEST_HARNESS)

#define CmdOutputBufferSize  128
unsigned char captureBuffer[CmdOutputBufferSize] = "test output for command execution\n";

static void captureStart(void)
{
  printf("starting cmd output capture\n");
}

static void captureStop(void)
{
  printf("stopping cmd output capture\n");
  captureIndex = strlen((char*)captureBuffer);
}

#else

#define CmdOutputBufferSize (128*1024)
#define captureBuffer ((uchar*)(TEXT_BASE-(2*CmdOutputBufferSize)))

static void capturePutc (const char c)
{
  if (captureIndex < CmdOutputBufferSize)
    captureBuffer[captureIndex++] = c;
}

static void capturePuts (const char *s)
{
  while (*s && (captureIndex < CmdOutputBufferSize))
    captureBuffer[captureIndex++] = *s++;
}

static void captureStart(void)
{
  printf("starting cmd output capture\n");
  static unsigned initDone;
  if (!initDone)
    {
      struct stdio_dev dev;
      memset(&dev, 0 , sizeof dev);
      strcpy(dev.name,"capture");
      dev.ext   = 0;
      dev.flags = DEV_FLAGS_OUTPUT;
      dev.putc  = capturePutc;
      dev.puts  = capturePuts;
      stdio_register(&dev);
      initDone = 1;
    }
  captureIndex = 0;
  console_assign(stdout,"capture");
  console_assign(stderr,"capture");
}

static void captureStop(void)
{
  console_assign(stdout,"serial");
  console_assign(stderr,"serial");
  printf("stopping cmd output capture (%d bytes)\n",captureIndex);
}
#endif

static void stringToLower(char *s)
{
  for (; *s; ++s)
    if (*s >= 'A' && *s <= 'Z')
      *s += 'a' - 'A';
}

static inline int str2long(char *p, ulong *num)
{
  char *endptr;
  *num = simple_strtoul(p, &endptr, 0);
  return (*p != '\0' && *endptr == '\0') ? 1 : 0;
}

static void parseAddressLength(const char *filename, unsigned long *addr, unsigned long *len)
{
#if defined(CONFIG_CMD_MTDPARTS)
  struct mtd_device *dev;
  struct part_info *part = NULL;
  u8 pnum;
#endif
  char *p = strchr(filename,'@');
  char *t = strchr(filename,':');

  if (t)
    *t = '\0';
  
  if (addr)
    {
      if (!p)
        *addr = 0xffffffff;
      else if (str2long(p+1,addr))
        ; // parsed OK
#if defined(CONFIG_CMD_MTDPARTS)
      else if ((mtdparts_init() == 0) && (find_dev_and_part(p+1, &dev, &pnum, &part) == 0)) 
        {
          // parsed OK as a partition name or device/partnum
          if (dev->id->type == MTD_DEV_TYPE_NAND && !strncmp(filename,"nand",4))
            *addr = part->offset;
          else
            {
              printf("Parition on wrong device\n");
              *addr = 0xffffffff;
            }
        }
#endif
      else
        *addr = 0xffffffff;
    }

  if (len)
    {
      if (!t)
        {
          // no len specified
#if defined(CONFIG_CMD_MTDPARTS)
          if (part)
            *len = part->size;
          else
#endif
            *len = 0xffffffff;
        }
      else if (!str2long(t+1,len))
        *len = 0xffffffff;
    }
}

//------------------------------------------------------------
// Structure used for varioius "memory drivers" used by tftp
// server.

typedef struct MemDriver_t
{
  const char *name;
  int (*open)(void);
  int (*get)(unsigned long blockNumber, uchar **bufptr, unsigned *lenptr);
  int (*put)(unsigned long blockNumber, uchar *bufptr, unsigned lenptr);
  void (*close)(void);
} MemDriver_t;

static char transferFilename[128];
static unsigned long transferOffset;
static unsigned long transferLength;
static unsigned transferOpcode;

#if EnableFakeMemoryDriver
//------------------------------------------------------------
// driver for "fake" RAM emulated by an in-program array
// $fakeram@addr:len
#define FakeMemorySize (4*1024)
static uchar fakeMemoryImage[FakeMemorySize];

static int FakeMemOpen(void)
{
  parseAddressLength(transferFilename,&transferOffset, &transferLength);
  printf("FakeMem open %lx:%lx opcode=%d\n",transferOffset, transferLength,transferOpcode);
  if (transferOffset >= sizeof fakeMemoryImage)
    return -1;
  if ((transferLength != 0xffffffff) && (transferOffset+transferLength > sizeof fakeMemoryImage))
    return -2;
  if ((transferLength == 0xffffffff) && (transferOpcode == TFTP_READ_REQ))
    return -3;
  return 0;
}
static int FakeMemGet(unsigned long blockNumber, uchar **bufptr, unsigned *lenptr)
{
  unsigned byteCount;
  if (transferOffset+(blockNumber*512) >= sizeof fakeMemoryImage)
    return -1;
  byteCount = transferLength - (blockNumber*512);
  if (byteCount > 512)
    byteCount = 512;
  *bufptr = fakeMemoryImage+transferOffset+(blockNumber*512);
  *lenptr = byteCount;
  return 0;
}
static int FakeMemPut(unsigned long blockNumber, uchar *buf, unsigned len)
{
  if (transferOffset+(blockNumber*512)+len >= sizeof fakeMemoryImage)
    return -1;
  memcpy(fakeMemoryImage+transferOffset+(blockNumber*512),buf,len);
  return 0;
}
#endif


#if EnableRamDriver
//------------------------------------------------------------
// $ram@addr:len

#define RamStart  0x20000000
#define RamEnd    TEXT_BASE


static int RamOpen(void)
{
  parseAddressLength(transferFilename,&transferOffset, &transferLength);
  printf("Ram open %lx:%lx opcode=%d\n",transferOffset, transferLength,transferOpcode);
  if ((transferOffset < RamStart) || (transferOffset >= RamEnd))
    return -1;
  if ((transferLength != 0xffffffff) && (transferOffset+transferLength > RamEnd))
    return -2;
  if ((transferLength == 0xffffffff) && (transferOpcode == TFTP_READ_REQ))
    return -3;
  return 0;
}
static int RamGet(unsigned long blockNumber, uchar **bufptr, unsigned *lenptr)
{
  unsigned byteCount;
  if (transferOffset+(blockNumber*512) >= RamEnd)
    return -1;
  byteCount = transferLength - (blockNumber*512);
  if (byteCount > 512)
    byteCount = 512;
  *bufptr = (uchar*)transferOffset+(blockNumber*512);
  *lenptr = byteCount;
  return 0;
}
static int RamPut(unsigned long blockNumber, uchar *buf, unsigned len)
{
  if (transferOffset+(blockNumber*512)+len >= RamEnd)
    return -1;
  memcpy((uchar*)transferOffset+(blockNumber*512),buf,len);
  return 0;
}
#endif

#if EnableNandDriver
#define NandBufferSize (128*1024)  // must be erase block size
#define NandWriteSize (4*1024)
#define NandBuffer ((uchar*)0x21800000) //TBD: fix this!

static int NandOpen(void)
{
  parseAddressLength(transferFilename, &transferOffset, &transferLength);
  printf("Nand open %lx:%lx opcode=%d\n",transferOffset, transferLength,transferOpcode);
  if (transferOpcode == TFTP_READ_REQ)
    {
      if (transferLength == 0xffffffff)
        return -1;
      if (transferOffset & (512-1))
        return -2;  // read must start on 512-byte boundary
    }
  if (transferOpcode == TFTP_WRITE_REQ)
    {
      if (transferOffset & (NandBufferSize-1))
        return -2;  // write must start on buffer-size boundary
    }

  return 0;
}
static int NandGet(unsigned long blockNumber, uchar **bufptr, unsigned *lenptr)
{
  unsigned byteCount;
  static unsigned bufferPresent, readOffset;
  unsigned bufferNeeded;

  if (blockNumber == 0)
    {
      bufferPresent = 0xffffffff;
      readOffset = transferOffset;
    }

  byteCount = transferLength - (blockNumber*512);

  if (byteCount)
    {
      if (byteCount > 512)
        byteCount = 512;
      
      // is requested block present in RAM buffer?
      bufferNeeded = (transferOffset +(blockNumber*512)) & ~(NandBufferSize-1);
      
      if (bufferNeeded != bufferPresent)
        {
          size_t len = NandBufferSize;
          int r;

          while (nand_block_isbad(&nand_info[0], readOffset))
            {
              printf("Skipping bad block 0x%08x\n", readOffset);
              readOffset += NandBufferSize;
            }

          len = NandBufferSize;

          r = nand_read(&nand_info[0], readOffset, &len, NandBuffer);
          if (r && (r != -EUCLEAN))
            {
              printf ("NAND read from offset %x failed %d\n", readOffset, r);
              return r;
            }

          readOffset += NandBufferSize;
          bufferPresent = bufferNeeded;
        }
      *bufptr = NandBuffer + ((transferOffset+(blockNumber*512)) & (NandBufferSize-1));
    }

  *lenptr = byteCount;
  return 0;
}

// Write a single block to flash, skipping bad blocks.  Return how
// many blocks to advance the offset (no bad blocks returns 1,
// skipping N bad blocks returns N+1

static int NandPutBlock(unsigned offset, uchar *buf, size_t len)
{
  unsigned skipped = 0;
  nand_erase_options_t opts;
  size_t wlen;

  while (1)
    {
      // skip bad blocks
      while (nand_block_isbad(&nand_info[0], offset))
        {
          printf("Skipping bad block 0x%08x\n", offset);
          offset += NandBufferSize;
          ++skipped;
        }

      // erase next block
      memset(&opts, 0, sizeof opts);
      opts.offset = offset;
      opts.length = NandBufferSize;
      
      if (nand_erase_opts(&nand_info[0], &opts))
        {
          printf("Error: nand_erase_opts() fail\n");
          return -1;
        }

      if (nand_block_isbad(&nand_info[0], offset))
        {
          printf("nand_erase_opts(offset=%08x) has marked block bad\n", offset);
          continue;
        }

      wlen = len;
      if (nand_write(&nand_info[0], offset, &wlen, buf))
        {
          printf("nand_write(offset=%08x, len=%d,%d) failed: marking block bad.\n", offset, len, wlen);
          nand_block_markbad(&nand_info[0], offset);
          continue;
        }
      
      break;
    }

  return skipped+1;
}


static int NandPut(unsigned long blockNumber, uchar *buf, unsigned len)
{
  static unsigned bufferPresent, writeOffset;
  unsigned bufferNeeded;
  uchar *bufptr;
  size_t byteCount;

  // is destination block present in RAM buffer?
  bufferNeeded = (transferOffset +(blockNumber*512)) & ~(NandBufferSize-1);

  if (blockNumber == 0)
    {
      bufferPresent = bufferNeeded;
      writeOffset = transferOffset;
    }
  
  // need to write old buffer to flash?
  if (bufferNeeded != bufferPresent)
    {
      int r = NandPutBlock(writeOffset, NandBuffer, NandBufferSize);
      if (r<0)
        return r;
      if (len == 0)
        return 0;
      writeOffset += r * NandBufferSize;
      bufferPresent = bufferNeeded;
    }

  bufptr = NandBuffer + ((blockNumber*512) & (NandBufferSize-1));
  memcpy(bufptr,buf,len);

  // last block if not a full packet
  if (len < 512)
    {
      byteCount = ((blockNumber*512) + len) & (NandBufferSize-1);
      if (byteCount)
        {
          // round up to multiple of NandWriteSize
          byteCount = (byteCount + (NandWriteSize-1)) & ~(NandWriteSize-1);
          if (NandPutBlock(writeOffset,NandBuffer,byteCount)<0)
            return -1;
        }
    }

  return 0;
}
static int NandEraseOpen(void)
{
  parseAddressLength(transferFilename,&transferOffset, &transferLength);
  printf("NandErase open %lx:%lx opcode=%d\n",transferOffset, transferLength,transferOpcode);
  if (transferLength == 0xffffffff)
    return -1;
  if (transferOffset & (NandBufferSize-1))
    return -2;  // erase must start on buffer-size boundary
  if (transferLength & (NandBufferSize-1))
    return -3;  // erase length must be multple of buffer-size

  nand_erase_options_t opts;
  memset(&opts, 0, sizeof opts);
  opts.offset = transferOffset;
  opts.length = transferLength;
  
  if (nand_erase_opts(&nand_info[0], &opts))
    return -4;
  return 0;
}
#endif

//------------------------------------------------------------
// $reset@
static void doReset(void)
{
  printf("tftpserver: resetting unit\n");
  do_reset();
}

//------------------------------------------------------------
// $go@addr
typedef void (*voidfp_t)(void);
static unsigned long GoAddress;
static int GoOpen(void)
{
  parseAddressLength(transferFilename,&GoAddress,NULL);
  if (GoAddress == 0xffffffff)
    return -1;
  return 0;
}
static void GoClose(void)
{
  voidfp_t fp = (voidfp_t)GoAddress;
  printf("tftpserver: go to address %p\n",fp);
  fp();
}

//------------------------------------------------------------
// $exitfail@
static void ExitFailClose(void)
{
  NetState = NETLOOP_FAIL;
}

//------------------------------------------------------------
// $exitok@
static void ExitOkClose(void)
{
  NetState = NETLOOP_SUCCESS;
}

//------------------------------------------------------------
// $cmd@<uboot-command-string>

static int CmdOpen(void)
{
  if (transferOpcode == TFTP_READ_REQ)
    {
      char *p = strchr(transferFilename,'@');
      if (p && p[1])
        {
          captureStart();
          run_command(p+1,0);
          captureStop();
        }
    }
  return 0;
}

static int CmdGet(unsigned long blockNumber, uchar **bufptr, unsigned *lenptr)
{
  unsigned byteCount;
  if (blockNumber*512 > captureIndex)
    return -1;
  byteCount = captureIndex - (blockNumber*512);
  if (byteCount > 512)
    byteCount = 512;
  *bufptr = captureBuffer+(blockNumber*512);
  *lenptr = byteCount;
  return 0;
}


static void CmdClose(void)
{
  if (transferOpcode == TFTP_WRITE_REQ)
    {
      char *p = strchr(transferFilename,'@');
      if (p && p[1])
        run_command(p+1,0);
    }
}

static const MemDriver_t MemoryDrivers[] =
  {
#if EnableFakeMemoryDriver
    {
      .name = "fakeram",
      .open = FakeMemOpen,
      .get  = FakeMemGet,
      .put  = FakeMemPut
    },
#endif
#if EnableRamDriver
    {
      .name  = "ram",
      .open  = RamOpen,
      .get   = RamGet,
      .put   = RamPut
    },
#endif
#if EnableNandDriver
    {
      .name  = "nand",
      .open  = NandOpen,
      .get   = NandGet,
      .put   = NandPut,
    },
    {
      .name  = "nanderase",
      .open  = NandEraseOpen,
    },
#endif
    {
      .name  = "reset",
      .close = doReset
    },
    {
      .name  = "go",
      .open  = GoOpen,
      .close = GoClose
    },
    {
      .name  = "exitfail",
      .close = ExitFailClose
    },
    {
      .name  = "exitok",
      .close = ExitOkClose
    },
    {
      .name  = "cmd",
      .open  = CmdOpen,
      .get   = CmdGet,
      .close = CmdClose
    },
    {
      .name = NULL,
    }
  };

static const MemDriver_t *lookupMemoryDriver(const char* filename)
{
  const MemDriver_t *m;
  const char *p;
  int n;

  if ( (p=strchr(filename,'@')) == NULL )
    return NULL;

  n = p-filename;

  for (m=MemoryDrivers; m->name; ++m)
    {
      if ((!strncmp(filename,m->name,n)) && strlen(m->name)==n)
        return m;
    }

  return NULL;
}

typedef enum{ServerIdle, ServerWrite, ServerRead, ServerReadDone} tftpState_t;

static tftpState_t serverState;

static unsigned otherPort;
static uchar otherEther[6];
static unsigned myPort = ListenPort;  // RFC1035 says it should be random;
static IPaddr_t otherIP;
static unsigned long blockNumber;
static unsigned long byteCount;
static unsigned timeoutCount;
static unsigned idleMax;
static unsigned subsequentIdleMax;

#define PutByte(p,b) *p++ = b
#define PutShort(p,s) do {*p++ = s>>8; *p++ = s;} while(0)
#define PutString(p,s) do {strcpy((char*)p,s); p += strlen(s)+1;} while (0)
#define PutBlock(p,b,len) do {memcpy(p,b,len); p += len;} while (0)

static void sendError(unsigned errorCode, char *errorMessage)
{
  uchar *pkt = (uchar*)NetTxPacket + NetEthHdrSize() + IP_HDR_SIZE;
  uchar *p = pkt;

  PutShort(p,TFTP_ERROR);
  PutShort(p,errorCode);
  PutString(p,errorMessage);
  printf("sendError %d '%s'\n",errorCode,errorMessage);
  NetSendUDPPacket(otherEther,otherIP,otherPort,myPort,(p-pkt));
}

static void sendAck(uint16_t blockNumber)
{
  uchar *pkt = (uchar*)NetTxPacket + NetEthHdrSize() + IP_HDR_SIZE;
  uchar *p = pkt;

  //printf("sendAck(%d)\n",blockNumber);
  PutShort(p,TFTP_ACK);
  PutShort(p,blockNumber);
  NetSendUDPPacket(otherEther,otherIP,otherPort,myPort,(p-pkt));
}

static int sendDataBlock(unsigned blockNumber, const MemDriver_t *mem)
{
  uchar *buf;
  unsigned len = 0;
  uchar *pkt = (uchar*)NetTxPacket + NetEthHdrSize() + IP_HDR_SIZE;
  uchar *p = pkt;
  PutShort(p,TFTP_DATA);
  PutShort(p,(uint16_t)blockNumber);
  if (mem && mem->get && mem->get(blockNumber-1,&buf,&len))
    {
      sendError(2,"Memory get failed");
      return -1;
    }
  PutBlock(p,buf,len);
  NetSendUDPPacket(otherEther,otherIP,otherPort,myPort,(p-pkt));
  //printf("sendData(%d,%p,%u)\n",blockNumber,buf,len);
  byteCount += len;
  return len;
}

// current memory driver
static const MemDriver_t *mem = NULL;

// handle receive timeouts
static void timeoutHandler (void)
{
  //printf("TFTP Server Timeout\n");
  NetSetTimeout(TIMEOUT * 1000, timeoutHandler);
  ++timeoutCount;

  if (serverState == ServerIdle)
    {
      if (idleMax && (timeoutCount >= idleMax))
        {
          printf("Idle timeout: tftp server terminating.\n");
          NetState = NETLOOP_SUCCESS;
        }
    }
  else
    {
      printf("timeout...\n");
      if (timeoutCount > TIMEOUTMAX)
        {
          printf("Too many timeouts, aborting transfer.\n");
          serverState = ServerIdle;
          return;
        }

      switch (serverState)
        {
        case ServerReadDone:
        case ServerRead:
          // waiting for ACK, so resend data
          sendDataBlock(blockNumber,mem);
          break;
        case ServerWrite:
          // waiting for data, so resent ACK
          sendAck((uint16_t)blockNumber);
          break;
        default:
          break;
        }
    }
}


// handle received tftp packets

static void handleTftpPacket(uchar *pkt, unsigned dest, unsigned src, unsigned len)
{
  char *mode = "";
  ushort opcode;

  NetSetTimeout(0, NULL);  // disable timeout
  timeoutCount = 0;

  opcode = (pkt[0]<<8) | pkt[1];

  //printf("State %d Opcode %d\n",serverState,opcode);

  switch (serverState)
    {
    case ServerIdle:
      // waiting for read/write request

      if (opcode==TFTP_READ_REQ || opcode==TFTP_WRITE_REQ)
        {
          IP_t *ip = (IP_t*)(pkt-IP_HDR_SIZE);
          otherPort = src;
          memcpy(otherEther,(pkt-IP_HDR_SIZE-NetEthHdrSize())+6,6);
          memcpy(&otherIP,&ip->ip_src,sizeof otherIP);
          mode = strchr((char*)&pkt[2],'\0')+1;
          strncpy(transferFilename,(char*)&pkt[2],sizeof transferFilename);
          transferFilename[(sizeof transferFilename)-1] = '\0';
          transferOpcode = opcode;
          idleMax = subsequentIdleMax;
          if (1)
            {
              printf("TFTP Server: rx %s request: '%s' mode '%s' ",
                     opcode==TFTP_READ_REQ ? "read" : "write",
                     transferFilename,mode);
              printf("from %lu.%lu.%lu.%lu:%d  ",(otherIP>>0)&0xff,(otherIP>>8)&0xff,(otherIP>>16)&0xff,(otherIP>>24)&0xff,otherPort);
              printf("[%02x:%02x:%02x:%02x:%02x:%02x]\n",otherEther[0],otherEther[1],otherEther[2],otherEther[3],otherEther[4],otherEther[5]);
            }
          stringToLower(mode);
          if (strcmp(mode,"octet"))
            {
              // we only support octet mode
              sendError(2,"Illegal trasfer mode. Only octet mode is supported.");
              break;
            }
          mem = lookupMemoryDriver(transferFilename);
          if (mem == NULL)
            {
              sendError(1,"Illegal memory/command name");
              break;
            }
          if (mem->open && mem->open())
            {
              sendError(3,"Memory open failed");
              break;
            }
          blockNumber = 0;
          byteCount = 0;
        }

      switch (opcode)
        {
        case TFTP_READ_REQ:
          {
            int status;
            blockNumber=1;
            status = sendDataBlock(blockNumber,mem);
            if (status == 512)
              serverState = ServerRead;
            else if (status >= 0)
              serverState = ServerReadDone;
            break;
          }
        case TFTP_WRITE_REQ:
          {
            serverState = ServerWrite;
            sendAck((uint16_t)blockNumber);
            break;
          }
        default:
          // invalid packet in this state
          sendError(4,"Illegal opcode when server in idle state.");
          break;
        }
      break;

    case ServerWrite:
      // waiting for data
      switch (opcode)
        {
        case TFTP_DATA:
          {
            unsigned datalen = len-4;
            uint16_t rxBlockNumber = (pkt[2]<<8) | pkt[3];

            //printf("rx block %u (%lu) %d bytes (%lu)\n",rxBlockNumber,blockNumber,datalen,byteCount);

            if ((uint16_t)rxBlockNumber != (uint16_t)(blockNumber+1))
              {
                sendAck((uint16_t)blockNumber);
                break;
              }

            ++blockNumber;

            // handle receive data
            if (mem->put && mem->put(blockNumber-1,&pkt[4],datalen))
              {
                sendError(3,"Memory put failed");
                serverState = ServerIdle;
              }

            sendAck((uint16_t)blockNumber);
            byteCount += datalen;

            if (datalen < 512)
              {
                // this is the last packet
                printf("done: received %lu bytes\n",byteCount);
                serverState = ServerIdle;
                if (mem && mem->close)
                  mem->close();
              }

          break;
          }

        default:
          // invalid packet in this state
          sendError(4,"Illegal opcode when server in write state.");
          break;
        }
      break;

    case ServerRead:
      // waiting for ACK
      switch (opcode)
        {
        case TFTP_ACK:
          {
            uint16_t rxBlockNumber = (pkt[2]<<8) | pkt[3];
            //printf("rx ACK %d\n",rxBlockNumber);
            if (rxBlockNumber == (uint16_t)blockNumber)
              ++blockNumber;
            if (sendDataBlock(blockNumber,mem) < 512)
              serverState = ServerReadDone;
            break;
          }
        default:
          // invalid packet in this state
          sendError(4,"Illegal opcode when server in read state.");
          break;
        }
      break;

    case ServerReadDone:
      // waiting for ACK of last packet
      switch (opcode)
        {
        case TFTP_ACK:
          {
            uint16_t rxBlockNumber = (pkt[2]<<8) | pkt[3];
            //printf("rx ACK %d\n",rxBlockNumber);
            if (rxBlockNumber != (uint16_t)blockNumber)
              sendDataBlock(blockNumber,mem);
            else
              {
                printf("done: sent %lu bytes\n",byteCount);
                serverState = ServerIdle;
                if (mem && mem->close)
                  mem->close();
              }
            break;
          }
        default:
          // invalid packet in this state
          sendError(4,"Illegal opcode when server in read state.");
          break;
        }
      break;

    default:
      // invalid state
      sendError(4,"Server in illegal state.");
      serverState = ServerIdle;
      break;
    }

  NetSetTimeout(TIMEOUT * 1000, timeoutHandler);
}

extern void handleDiscoveryPacket(uchar *pkt, unsigned dest, unsigned src, unsigned len);

// handle received UDP packets

static void rxHandler (uchar *pkt, unsigned dest, unsigned src, unsigned len)
{
  //printf("Tftp Server rxHandler() got packet from port %u to port %u len %d\n",src,dest,len);

  if (dest == ListenPort && len >= 4)
    handleTftpPacket(pkt, dest, src, len);
  else if (dest == 4606)
    handleDiscoveryPacket(pkt, dest, src, len);
}

void TftpServerStart (void)
{
  printf("Starting TFTP server on port %d\n", ListenPort);

  serverState = ServerIdle;
  blockNumber = 0;
  timeoutCount = 0;
  NetSetTimeout(TIMEOUT * 1000, timeoutHandler);
  NetSetHandler(rxHandler);
}

static int do_tftpserver (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
  if (argc>=2)
    idleMax = ((simple_strtoul(argv[1],NULL,0) * 100)/72);
  else
    idleMax = 0;

  if (argc==3)
    subsequentIdleMax = ((simple_strtoul(argv[2],NULL,0) * 100)/72);
  else
    subsequentIdleMax = 0;

  return NetLoop(TFTPSRV) == -1;
}

U_BOOT_CMD(tftpserver,3,0,do_tftpserver,"start TFTP server","[initial-timeout [timeout]]");

#endif
#endif



// This file uses different indentation than the rest of the U-Boot
// source files, this is how let emacs know
//
// Local Variables:
// c-file-style: "gnu"
// End:
